/**
 * \file Basic_Consumer.hh
 * \author Scott F. H. Kaplan -- sfkaplan@cs.amherst.edu
 * \date June 2004
 * \brief Produces a single reference stream to canonical page IDs.
 *
 * A subclass of Consumer that emits a basic page-level reference
 * trace that reflects the referencing of the total original workload,
 * with no separation by tasks or processes.  Canonical page numbers
 * are emitted so that shared pages are not multiply counted.  The
 * trace takes the following form:
 *
 *   R 14fe3
 *
 * The first field is the ``type tag'', which is an 8-bit character
 * that takes one of the following possible values:
 *
 *   -> 'r' (single virtual memory read)
 *   -> 'R' (multiple virtual memory read)
 *   -> 'w' (single virtual memory write)
 *   -> 'W' (multiple virtual memory write)
 *   -> '<' (file system read)
 *   -> '>' (file system write)
 *
 * The second field is the page number: A 32-bit hexidecimal integer
 * giving the canonical page number being referenced.
 **/



// ===================================================================
// Avoid multiple inclusion.
#if !defined (_BASIC_CONSUMER_HH)
#define _BASIC_CONSUMER_HH
// ===================================================================



// ===================================================================
// INCLUDES

#include <zlib.h>
#include "Consumer.hh"
// ===================================================================



// ===================================================================
// CLASS DEFINITION

/**
 * \brief A subclass of Consumer that produces a simple reference
 *        trace.
 *
 * It relies on Consumer to reconcile virtual pages to canonical
 * pages, and then emits a string of references to those canonical
 * pages.
 **/
class Basic_Consumer : public Consumer {

public: // CONSTRUCTOR

  /**
   * \brief The constructor.
   * \param reader The object that will read and parse the reference
   *        and kernel traces.
   * \param output_file_pathname The pathname for the file that will
   *        contain the resultant reference stream.
   **/
  Basic_Consumer (Reader* const reader,
		  const char* const output_file_pathname);

  /**
   * \brief The deconstructor.
   *
   * Close the output file.
   **/
  virtual ~Basic_Consumer ();

protected: // VIRTUAL METHODS

  /** \brief Called after processing ends -- Does nothing. */
  virtual void clean_up () {}

  /**
   * \brief Called after processing a reference record.
   *
   * Emit the reference, using the canonical page ID, to the output
   * trace.
   **/
  virtual void act_on_reference_record ();

  /**
   * \brief Called after processing a kernel record.
   *
   * If the record indicates a file system read or write, emit the
   * reference, using the canonical page ID, to the output trace.
   **/
  virtual void act_on_kernel_record ();

protected: // DATA MEMBERS

  /** \brief The output file stream. */
  gzFile output_stream;

}; // class Basic_Consumer
// ===================================================================



// ===================================================================
#endif // _BASIC_CONSUMER_HH
// ===================================================================
