// ===================================================================
// Basic_Named_Consumer.cc
// Scott F. H. Kaplan -- sfkaplan@cs.amherst.edu
// March 2005

// A subclass of Basic_Consumer that emits a basic page-level
// reference trace for virtual memory and file system references,
// emitting canonical page IDs with tags that indicate dirtiness and
// the named/unnamed status of the page.  See the header file for more
// details.
// ===================================================================



// ===================================================================
// INCLUDES

#include <assert.h>
#include <fcntl.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include "Basic_Named_Consumer.hh"
// ===================================================================



// ===================================================================
Basic_Named_Consumer::Basic_Named_Consumer
(Reader* const reader,
 const char* const output_file_pathname) :
  Basic_Consumer(reader, output_file_pathname) {

  // Nothing additional to do.

} // Basic_Named_Consumer::Basic_Named_Consumer
// ===================================================================



// ===================================================================
void
Basic_Named_Consumer::act_on_reference_record () {

  // If we're still waiting to see the first scheduling event (which
  // would tell us which task is active), then we cannot yet process
  // this record.
  assert(merge_phase != PHASE_PRE_REFERENCE);
  if (merge_phase != PHASE_PROCESSING_REFERENCES) {
    return;
  }

  // Grab the virtual->canonical map for the given context.
  Process* process =
    safe_context_ID_to_process(reference_record->virtual_page.context_ID);
  virtual_to_canonical_map_t* map = process->virtual_to_canonical_map;

  // Lookup the canonical page.
  canonical_page_s* canonical_page =
    safe_virtual_to_canonical_page(map,
				   reference_record->virtual_page.page_ID);

  // Determine the tag to emit.  For onymous (file) pages, the tag
  // will stay the same; for anonymous (swap) pages, we change the tag
  // to record its anonymity.
  tag_t tag = reference_record->tag;
  if (canonical_page->type != TYPE_FILE) {

    switch (tag) {

    case TAG_SINGLE_READ:
      tag = TAG_SINGLE_ANONYMOUS_READ;
      break;

    case TAG_MULTIPLE_READ:
      tag = TAG_MULTIPLE_ANONYMOUS_READ;
      break;

    case TAG_SINGLE_WRITE:
      tag = TAG_SINGLE_ANONYMOUS_WRITE;
      break;

    case TAG_MULTIPLE_WRITE:
      tag = TAG_MULTIPLE_ANONYMOUS_WRITE;
      break;

    default:
      fprintf(stderr,
	      "Basic_Named_Consumer::act_on_reference_record(): ");
      fprintf(stderr,
	      "Unknown tag %c(%d) on anonymous page\n",
	      tag,
	      (int)tag);
      exit(1);

    }

  }

  // Emit the new record.
  int tagResult = gzwrite(output_stream, &tag, sizeof(tag_t));
  int pageIDResult = gzwrite(output_stream,
			     &canonical_page->page,
			     sizeof(canonical_page_ID_t));
  if ((tagResult == 0) || (pageIDResult == 0)) {
    fprintf(stderr, "Basic_Consumer::act_on_reference_record(): ");
    fprintf(stderr, "Failed record write\n");
    exit(1);
  }

} // act_on_reference_record
// ===================================================================



// ===================================================================
void
Basic_Named_Consumer::act_on_kernel_record () {

  // Only emit a record for a file system read or write operation, and
  // emit those only if reference processing has begun.
  if ((merge_phase == PHASE_PROCESSING_REFERENCES) &&
      ((kernel_record->tag == TAG_FILE_READ) ||
       (kernel_record->tag == TAG_FILE_WRITE))) {

    // Grab the virtual->canonical map for the given file.
    virtual_to_canonical_map_t* map =
      safe_file_ID_to_V2C_map(kernel_record->file_ID);

    // Loop through the file pages referenced by this operation.
    for (virtual_page_ID_t virtual_page_ID =
	   kernel_record->starting_file_page;
	 virtual_page_ID < kernel_record->ending_file_page;
	 virtual_page_ID++) {

      // Lookup the canonical page ID.
      canonical_page_ID_t canonical_page_ID =
	lookup_canonical_page_ID(map, virtual_page_ID);

      // Emit the new record.
      int tagResult = gzwrite(output_stream,
			      &kernel_record->tag,
			      sizeof(tag_t));
      int pageIDResult = gzwrite(output_stream,
				 &canonical_page_ID,
				 sizeof(canonical_page_ID_t));
      if ((tagResult == 0) || (pageIDResult == 0)) {
	fprintf(stderr, "Basic_Named_Consumer::act_on_kernel_record():");
	fprintf(stderr, "Failed record write\n");
	exit(1);
      }

    }
  }
} // act_on_kernel_record
// ===================================================================
