/**
 * \file Basic_Named_Consumer.hh
 * \author Scott F. H. Kaplan -- sfkaplan@cs.amherst.edu
 * \date March 2005
 * \brief Produces a single reference stream to canonical page IDs
 *        with tags indicating dirtiness and named status.
 *
 * A subclass of Basic_Consumer that emits a basic page-level
 * reference trace that reflects the referencing of the total original
 * workload, with no separation by tasks or processes.  Canonical page
 * numbers are emitted so that shared pages are not multiply counted.
 * The output format is in binary, and each record contains two
 * fields:
 *
 *   1) A ``type tag'', which is an 8-bit character that takes one of
 *      the following possible values:
 *
 *      -> 'l' (single virtual memory read of anonymous page)
 *      -> 'L' (multiple virtual memory read of anonymous page)
 *      -> 's' (single virtual memory write of anonymous page)
 *      -> 'S' (multiple virtual memory write of anonymous page)
 *      -> 'r' (single virtual memory read of onymous page)
 *      -> 'R' (multiple virtual memory read of onymous page)
 *      -> 'w' (single virtual memory write of onymous page)
 *      -> 'W' (multiple virtual memory write of onymous page)
 *      -> '<' (file system read)
 *      -> '>' (file system write)
 *
 *   2) A page number: A 32-bit integer giving the canonical page
 *      number being referenced.  A key difference between this class
 *      and its parent is the distinction between anonymous and
 *      onymous pages.
 **/



// ===================================================================
// Avoid multiple inclusion.
#if !defined (_BASIC_NAMED_CONSUMER_HH)
#define _BASIC_NAMED_CONSUMER_HH
// ===================================================================



// ===================================================================
// INCLUDES

#include <zlib.h>
#include "Basic_Consumer.hh"
// ===================================================================



// ===================================================================
// CLASS DEFINITION

/**
 * \brief A subclass of Basic_Consumer that produces a simple
 *        reference trace.
 *
 * It relies on Consumer to reconcile virtual pages to canonical
 * pages, and then emits a string of references to those canonical
 * pages.
 **/
class Basic_Named_Consumer : public Basic_Consumer {

public: // CONSTRUCTOR

  /**
   * \brief The constructor.
   * \param reader The object that will read and parse the reference
   *        and kernel traces.
   * \param output_file_pathname The pathname for the file that will
   *        contain the resultant reference stream.
   **/
  Basic_Named_Consumer (Reader* const reader,
			const char* const output_file_pathname);

protected: // VIRTUAL METHODS

  /**
   * \brief Called after processing a reference record.
   *
   * Emit the reference, using the canonical page ID, to the output
   * trace.
   **/
  virtual void act_on_reference_record ();

  /**
   * \brief Called after processing a kernel record.
   *
   * If the record indicates a file system read or write, emit the
   * reference, using the canonical page ID, to the output trace.
   **/
  virtual void act_on_kernel_record ();

}; // class Basic_Named_Consumer
// ===================================================================



// ===================================================================
// Valid reference trace tags specific to this consumer type.
#define TAG_SINGLE_ANONYMOUS_READ    'l'
#define TAG_MULTIPLE_ANONYMOUS_READ  'L'
#define TAG_SINGLE_ANONYMOUS_WRITE   's'
#define TAG_MULTIPLE_ANONYMOUS_WRITE 'S'
// ===================================================================



// ===================================================================
#endif // _BASIC_NAMED_CONSUMER_HH
// ===================================================================
