/**
 * \file Consumer.cc
 * \author Scott F. H. Kaplan <sfkaplan@cs.amherst.edu>
 * \date June 2004
 * \brief Code of base class for reading traces and rebuilding system
 *        state.
 *
 * A class used to read both a kernel and reference trace and, with
 * each record, update data structures that reflect the state of the
 * processes and the memory.
 *
 * Note that this is a PURE VIRTUAL CLASS: It does not define the
 * act_on_kernel_record(), act_on_reference_record(), and clean_up()
 * methods.  By defining these methods, a subclass can choose to
 * maintain extra data structures or to produce output before
 * processing, at the processing of each record, and after processing.
 **/



// ===================================================================
// INCLUDES

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "Consumer.hh"
// ===================================================================



// ===================================================================
Consumer::Consumer (Reader* const reader) {

  // Hold a pointer to the reference reader.
  this->reader = reader;

  // Initialize some data members.  Start the canonical page ID's at
  // 1, since some simulators may reasonably barf on a page number 0.
  // We also start at 1 so that we can easily detect wrap-around.
  next_canonical_page_ID = 1;
  canonical_zero_page = get_new_canonical_page(TYPE_ANONYMOUS);
  next_process_ID = KERNEL_PROCESS_ID + 1;
  merge_phase = PHASE_PRE_REFERENCE;

  // Create space to hold one reference record and one kernel record.
  reference_record = new reference_record_s;
  kernel_record = new kernel_record_s;
  if ((reference_record == 0) || (kernel_record == 0)) {
    fprintf(stderr, "Consumer(): Unable to allocate record space\n");
    exit(1);
  }

  // Allocate and initialize the the handle dispatch table, then fill
  // it with the correct entries.
  handler_dispatch_table = new handler_t[NUMBER_POSSIBLE_TAG_VALUES];
  if (handler_dispatch_table == NULL) {
    fprintf(stderr, "Consumer::Consumer(): ");
    fprintf(stderr, "Could not allocate handler dispatch table\n");
    exit(1);
  }
  for (int i = 0; i < NUMBER_POSSIBLE_TAG_VALUES; i++) {
    handler_dispatch_table[i] = &Consumer::handle_invalid;
  }

  assign_handler(TAG_SCHEDULE, &Consumer::handle_schedule);
  assign_handler(TAG_FORK, &Consumer::handle_fork);
  assign_handler(TAG_EXEC, &Consumer::handle_exec);
  assign_handler(TAG_EXIT, &Consumer::handle_exit);
  assign_handler(TAG_CONTEXT_ASSIGNMENT, &Consumer::handle_context_assignment);
  assign_handler(TAG_DUPLICATE_RANGE, &Consumer::handle_duplicate_range);
  assign_handler(TAG_MMAP_FILE, &Consumer::handle_file_mmap_range);
  assign_handler(TAG_MMAP_ANONYMOUS, &Consumer::handle_anonymous_mmap_range);
  assign_handler(TAG_MREMAP, &Consumer::handle_mremap);
  assign_handler(TAG_MUNMAP, &Consumer::handle_munmap_range);
  assign_handler(TAG_COMPLETE_UNMAP, &Consumer::handle_complete_unmap);
  assign_handler(TAG_COW_UNMAP, &Consumer::handle_cow_unmap);
  assign_handler(TAG_SHMAT, &Consumer::handle_shmat);
  assign_handler(TAG_SHM_DESTROY, &Consumer::handle_shm_destroy);
  assign_handler(TAG_FILE_OPEN, &Consumer::handle_no_op);
  assign_handler(TAG_FILE_CLOSE, &Consumer::handle_no_op);
  assign_handler(TAG_FILE_READ, &Consumer::handle_file_read_or_write);
  assign_handler(TAG_FILE_WRITE, &Consumer::handle_file_read_or_write);
  assign_handler(TAG_FILE_DELETE, &Consumer::handle_file_delete);
  assign_handler(TAG_ACCEPT, &Consumer::handle_no_op);
  assign_handler(TAG_INVALID, &Consumer::handle_invalid);
  assign_handler(TAG_END_OF_TRACE, &Consumer::handle_invalid);
  assign_handler(TAG_DEBUG, &Consumer::handle_cow_unmap);

  // Create the input buffer.
  buffer = new char[BUFFER_SIZE];
  if (buffer == NULL) {
    fprintf(stderr, "Consumer::Consumer(): ");
    fprintf(stderr, "Unable to allocation buffer\n");
    exit(1);
  }

} // Consumer::Consumer
// ===================================================================



// ===================================================================
void
Consumer::assign_handler (tag_t tag, handler_t handler) {

  int index = (int)tag;
  if (index >= NUMBER_POSSIBLE_TAG_VALUES) {
    fprintf(stderr,
	    "Consumer::assign_handler():  Invalid tag/index = %c/%d >= %d\n",
	    tag,
	    index,
	    NUMBER_POSSIBLE_TAG_VALUES);
    exit(1);
  }
  assert(handler_dispatch_table != NULL);
  handler_dispatch_table[index] = handler;

}
// ===================================================================



// ===================================================================
Consumer::~Consumer () {

  // WARNING: We do nothing here to clean up this object.  We (lazily)
  // assume that when this object dies, the process is complete
  // anyway.

} // Consumer::~Consumer
// ===================================================================



// ===================================================================
// Perform the merging.
void
Consumer::merge () {

  // Complete initialization.
  initialize();

  // Read the first kernel and reference records to be processed.
  reader->read_kernel_record(kernel_record);
  reader->read_reference_record(reference_record);

  // Merge the traces until the last reference has been processed.
  while (reference_record->tag != TAG_END_OF_TRACE) {

    // It should be impossible for the timestamps of the current
    // kernel and reference records to match.
    if (kernel_record->cycle_timestamp ==
	reference_record->cycle_timestamp) {

      fprintf(stderr, "Consumer::merge(): Identical timestamps\n");
      fprintf(stderr,
	      "  kernel_record->cycle_timestamp = %qx at record %qd\n",
	      kernel_record->cycle_timestamp,
	      kernel_record->record_number);
      fprintf(stderr,
	      "  reference_record->cycle_timestamp = %qx at record %qd\n",
	      reference_record->cycle_timestamp,
	      reference_record->record_number);
      exit(1);

    }

    // Which trace contains the next record?
    if (kernel_record->cycle_timestamp <
	reference_record->cycle_timestamp) {

      // The kernel trace record is next. Update the time that has
      // passed since the last record for the scheduled task.
      //
      // Special case: The previous record may have terminated the
      // scheduled task, thus obviating the need to mark its time.
      // However, be sure that the current record is a scheduling
      // record that will set a new scheduled task.
      if (scheduled_task == NULL) {

	assert(kernel_record->tag == TAG_SCHEDULE);

      } else {

	scheduled_task->mark_time(kernel_record->cycle_timestamp,
				  kernel_record->reference_timestamp);

      }

      // Update the maps with this kernel record, and act on the
      // record (that is, output and other bookkeeping).
      update_with_kernel_record();
      act_on_kernel_record();

      // Read the next kernel record.
      reader->read_kernel_record(kernel_record);
      
    } else {

      // The reference record is processed next.  Update the time that
      // has passed since the last record for the scheduled task.
      scheduled_task->mark_time(reference_record->cycle_timestamp,
				reference_record->reference_timestamp);

      // Update the maps with this reference record, and act on the
      // record (that is, output and other bookkeeping).
      update_with_reference_record();
      act_on_reference_record();

      // Read the next reference record.
      reader->read_reference_record(reference_record);

    }

  } // Record reading loop.

  // Complete processing by finishing output and other bookkeeping.
  clean_up();

} // Consumer::merge
// ===================================================================


// ===================================================================
Task*
Consumer::safe_task_ID_to_task (const task_ID_t task_ID) {

  // Attempt lookup.
  task_ID_to_Task_map_t::iterator i = task_ID_to_Task_map.find(task_ID);

  // If the lookup failed then abort.
  if (i == task_ID_to_Task_map.end()) {
    fprintf(stderr, "Consumer::safe_task_ID_to_task(): ");
    fprintf(stderr,
	    "Failed lookup at record %qd\n",
	    kernel_record->record_number);
    exit(1);
  }

  // Return the lookup result.
  return (*i).second;

}
// ===================================================================



// ===================================================================
Process*
Consumer::safe_context_ID_to_process (const context_ID_t context) {

  // Attempt lookup.
  context_ID_to_Process_map_t::iterator i =
    context_ID_to_Process_map.find(context);

  // If the lookup failed then abort.
  if (i == context_ID_to_Process_map.end()) {
    fprintf(stderr, "Consumer::safe_context_ID_to_process(): ");
    fprintf(stderr, "Failed lookup\n");
    exit(1);
  }

  // Return the lookup result.
  return (*i).second;

}
// ===================================================================



// ===================================================================
bool
Consumer::safe_live (const task_ID_t task_ID) {

  // Attempt lookup.
  live_map_t::iterator i = live_map.find(task_ID);

  // If the lookup failed then abort.
  if (i == live_map.end()) {
    fprintf(stderr, "safe_live(): Failed lookup\n");
    exit(1);
  }

  // Return the lookup result.
  return (*i).second;

}
// ===================================================================



// ===================================================================
virtual_to_canonical_map_t*
Consumer::safe_file_ID_to_V2C_map (const file_ID_s& file_ID) {

  virtual_to_canonical_map_t* map = 0;

  // Attempt lookup.
  file_ID_to_V2C_map_t::iterator i = file_ID_to_V2C_map.find(file_ID);

  // Did the lookup succeed?
  if (i == file_ID_to_V2C_map.end()) {

    // No.  Allocate a new map, and ensure that the allocation
    // succeeded.
    map = new virtual_to_canonical_map_t;
    if (map == 0) {
      fprintf(stderr, "safe_file_ID_to_V2C_map(): Failed allocation\n");
      exit(1);
    }

    // Associate the given file ID with this new map.
    file_ID_to_V2C_map[file_ID] = map;

  } else {

    // Yes, use the map found via lookup.
    map = (*i).second;

  }

  // Return the lookup result.
  return map;

}
// ===================================================================



// ===================================================================
virtual_to_canonical_map_t*
Consumer::safe_shm_ID_to_V2C_map (const shm_ID_t& shm_ID) {

  virtual_to_canonical_map_t* map = 0;

  // Attempt lookup.
  shm_ID_to_V2C_map_t::iterator i = shm_ID_to_V2C_map.find(shm_ID);

  // Did the lookup succeed?
  if (i == shm_ID_to_V2C_map.end()) {

    // No.  Allocate a new map, and ensure that the allocation
    // succeeded.
    map = new virtual_to_canonical_map_t;
    if (map == 0) {
      fprintf(stderr, "safe_shm_ID_to_V2C_map(): Failed allocation\n");
      exit(1);
    }

    // Associate the given shm ID with this new map.
    shm_ID_to_V2C_map[shm_ID] = map;

  } else {

    // Yes, use the map found via lookup.
    map = (*i).second;

  }

  // Return the lookup result.
  return map;

} // safe_shm_ID_to_V2C_map
// ===================================================================



// ===================================================================
canonical_page_s*
Consumer::safe_virtual_to_canonical_page
(const virtual_to_canonical_map_t* const map,
 const virtual_page_ID_t virtual_page_ID) {

  canonical_page_s* canonical_page;

  // Attempt to lookup the page in the context's map.
  const virtual_to_canonical_map_t::const_iterator i =
    (*map).find(virtual_page_ID);

  // Was the lookup successful?
  if (i != (*map).end()) {

    // Yes.  Grab a pointer to the page..
    canonical_page = (*i).second;

  } else {

    // There was no canonical page found.  Therefore, this is an
    // invalid virtual page ID for this context.  Emit the error and
    // exit.
    fprintf(stderr, "Consumer::safe_virtual_to_canonical_page(): ");
    fprintf(stderr, "Failed lookup\n");
    exit(1);

  }

  return canonical_page;

} // Consumer::safe_virtual_to_canonical_page
// ===================================================================



// ===================================================================
void
Consumer::map_virtual_to_canonical
(virtual_to_canonical_map_t* const map,
 const virtual_page_ID_t virtual_page_ID,
 canonical_page_s* const canonical_page) {

  // The local page about to be assigned should exist.
  // created.
  if (canonical_page == NULL) {
    fprintf(stderr,
	    "map_virtual_to_canonical(): Null canonical page\n");
    exit(1);
  }

  // Insert the mapping.
  (*map)[virtual_page_ID] = canonical_page;

  // Increase the mapping count.
  canonical_page->mapping_count++;

} // map_virtual_to_canonical
// ===================================================================



// ===================================================================
void
Consumer::delete_task (Task* task) {

  assert(task != NULL);
  delete task;

}
// ===================================================================



// ===================================================================
canonical_page_s*
Consumer::get_new_canonical_page (const page_type_t type) {

  // Ensure that we haven't exhausted the canonical ID space.  If the
  // next ID is 0, we've wrapped around.
  if (next_canonical_page_ID == 0) {
    fprintf(stderr,
	    "get_new_canonical_page(): Canonical page IDs exhausted\n");
    exit(1);
  }

  // Allocate a new canonical page structure.
  canonical_page_s* page = new canonical_page_s;
  if (page == 0) {
    fprintf(stderr, "get_new_canonical_page(): Allocation failed\n");
    exit(1);
  }

  // Assign the fields and return a pointer to it.  Assume initially
  // that there are no maps pointing to this page.
  page->page = next_canonical_page_ID++;
  page->mapping_count = 0;
  page->type = type;
  return page;

} // get_new_canonical_page
// ===================================================================



// ===================================================================
// Take a handle to the canonical page, making it possible to nullify
// the pointer to the page if it is deleted by this method.

void
Consumer::delete_if_unused_page
(canonical_page_s** const canonical_page_handle) {

  // Is this a page that is no longer mapped by any context?
  if ((*canonical_page_handle)->mapping_count == 0) {

    canonical_page_s* canonical_page = *canonical_page_handle;

    // If this page is anonymous, then delete it because it can never
    // be mapped again.
    if (canonical_page->type == TYPE_ANONYMOUS) {

      delete canonical_page;
      *canonical_page_handle = NULL;

    }

  }
} // Consumer::delete_if_unused_page
// ===================================================================



// ===================================================================
// Take a handle to the process, making it possible to nullify the
// pointer to the process if it is deleted by this method.

void
Consumer::delete_if_defunct_process (Process** const process_handle) {

  Process* process = *process_handle;

  // Are there any tasks left for this process?
  if (process->task_list.empty()) {

    // None.  Unmap any remaining pages in the context.
    perform_complete_unmap(process->context_ID);

    // Find and erase the context_ID->process mapping.
    context_ID_to_Process_map_t::iterator i =
      context_ID_to_Process_map.find(process->context_ID);
    if (i != context_ID_to_Process_map.end()) {

      context_ID_to_Process_map.erase(i);

    }

    // Delete the virtual->canonical map and the process object.
    delete process->virtual_to_canonical_map;
    delete process;
    *process_handle = 0;

  } // Process has no remaining tasks

} // Consumer::delete_if_defunct_process
// ===================================================================



// ===================================================================
void
Consumer::unlink_task_and_process (Task* const task) {

  // Grab a pointer the task's process.
  Process* process = task->process;

  // If the task is not associated with a process, then there's no
  // work to do here.
  if (process == NULL) {
    return;
  }

  // Clear the process pointer in the task object, thus performing
  // the first half of the unlinking.
  task->process = NULL;

  // Search through the process's task list for this task.
  list<Task*>::iterator i = process->task_list.begin();
  bool done = false;
  while (!done) {

    // Sanity check:  We should never reach the end of the list
    // without finding the task.
    if (i == process->task_list.end()) {
      fprintf(stderr, "Consumer::unlink_task_and_process(): ");
      fprintf(stderr, "Task not found in process\n");
      exit(1);
    }

    // If we've found it, delete it and end the search.  Otherwise,
    // keep searching.
    if (*i == task) {

      // It is.  Erase it from the list, thus performing the second
      // half of the unlinking.
      process->task_list.erase(i);
      done = true;

    } else {

      // It is not.  Continue searching.
      i++;

    }
  } // Loop over task IDs in the process

  // The process may be defunct as a result of the unlinking, and if
  // so, it should be deleted.
  delete_if_defunct_process(&process);

} // Consumer::unlink_task_and_process
// ===================================================================



// ===================================================================
void
Consumer::clear_address_space (virtual_to_canonical_map_t* map,
			       page_type_t type) {

  // Traverse every page in the map, deleting the underlying canonical
  // pages.
  virtual_to_canonical_map_t::iterator i = (*map).begin();
  while (i != (*map).end()) {

    // Grab the canonical page and decrement the count of mappings to
    // it (since this address space will no longer map it).
    canonical_page_s* canonical_page = (*i).second;
    canonical_page->mapping_count--;

    // Sanity check: The canonical page should now have a reference
    // count of zero, or else the system deleted a page that still had
    // a valid mapping.  The canonical page should also have the
    // requested type.
    assert(canonical_page->mapping_count == 0);
    assert(canonical_page->type == type);
    delete canonical_page;
    canonical_page = 0;

    // Move to the next mapped page.
    i++;

  }

} // Consumer::clear_address_space
// ===================================================================



// ===================================================================
void
Consumer::handle_schedule () {

  // Find the task object based on the given ID.
  scheduled_task =
    safe_task_ID_to_task(kernel_record->task_ID);

  // Mark the start of the quanta for the scheduled task.  Note that
  // reference time, provided as a virtual value, cannot have changed
  // for this task, but that cycle time, provided as a global value,
  // will have.
  scheduled_task->mark_quanta_start(kernel_record->cycle_timestamp);

  // If this is the first scheduling event since the first observed
  // reference, mark the change of phase to processing references.
  if (merge_phase == PHASE_WAITING_FOR_SCHEDULE) {
    merge_phase = PHASE_PROCESSING_REFERENCES;
  }

} // Consumer::handle_schedule
// ===================================================================



// ===================================================================
void
Consumer::handle_fork () {

  // Sanity check: There should be no task already using the given ID.
  task_ID_to_Task_map_t::iterator i =
    task_ID_to_Task_map.find(kernel_record->task_ID);
  if (i != task_ID_to_Task_map.end()) {
    fprintf(stderr, "Consumer::handle_fork(): Duplicate task ID\n");
    exit(1);
  }

  // Create a new task object.
  Task* child_task = allocate_new_task();

  // Look up the parent task.
  Task* parent_task =
    safe_task_ID_to_task(kernel_record->parent_task_ID);

  // Assign the new task's fields.  Initially, the task has the same
  // context as the parent.
  child_task->task_ID = kernel_record->task_ID;
  child_task->parent_task = parent_task;
  assert(parent_task->process != NULL);
  parent_task->process->task_list.push_front(child_task);
  child_task->process = parent_task->process;

  // Create a table entry mapping the task ID to the new task object.
  task_ID_to_Task_map[kernel_record->task_ID] = child_task;

} // Consumer::handle_fork
// ===================================================================



// ===================================================================
void
Consumer::handle_exec () {

  // Grab the task process associated with this task.
  Task* task = safe_task_ID_to_task(kernel_record->task_ID);
  Process* process = task->process;

  // Assign the new executable image.
  process->image_file_ID = kernel_record->file_ID;
  strcpy(process->image_pathname, kernel_record->filename);

} // Consumer::handle_exec
// ===================================================================



// ===================================================================
void
Consumer::handle_exit () {

  // Lookup the task and process objects for the given task ID.
  task_ID_to_Task_map_t::iterator task_iterator =
    task_ID_to_Task_map.find(kernel_record->task_ID);
  if (task_iterator == task_ID_to_Task_map.end()) {
    fprintf(stderr, "Consumer::handle_exit(): Failed lookup\n");
    exit(1);
  }
  Task* task = (*task_iterator).second;

  // Separate the task and its process.  If the process becomes
  // defunct as a result, it will be deleted.
  unlink_task_and_process(task);

  // Clear the task_ID->process mapping.
  task_ID_to_Task_map.erase(task_iterator);

  // Delete the task.  If this is the currently scheduled task, record
  // that fact too.
  if (task == scheduled_task) {
    scheduled_task = NULL;
  }
  delete_task(task);
  task = NULL;

} // Consumer::handle_exit
// ===================================================================



// ===================================================================
void
Consumer::handle_context_assignment () {

  // Grab the task object for the given task ID.
  Task* task = safe_task_ID_to_task(kernel_record->task_ID);

  // Unlink this task from its current process (if any).
  unlink_task_and_process(task);

  // Get the mapping from the given context ID to its process.  Is it
  // mapped to any process?
  context_ID_to_Process_map_t:: iterator process_iterator =
    context_ID_to_Process_map.find(kernel_record->context_ID);
  Process* process;
  if (process_iterator == context_ID_to_Process_map.end()) {

    // The context ID is not yet associated with a process.  Allocate
    // a new process object.
    process = allocate_new_process();

    // Assign the process a new ID.
    process->process_ID = next_process_ID++;
    if (process->process_ID == 0) {
      fprintf(stderr, "Consumer::handle_context_assignment(): ");
      fprintf(stderr, "Exhausted process ID space\n");
      exit(1);
    }

    // Create a new context by creating a new virtual->canonical map,
    // and assign it to the new process.
    process->virtual_to_canonical_map = new virtual_to_canonical_map_t;
    if (process->virtual_to_canonical_map == NULL) {
      fprintf(stderr, "Consumer::handle_context_assignment(): ");
      fprintf(stderr, "Failed allocation of virtual->canonical map\n");
      exit(1);
    }

    // Assign the given context ID to the process.
    process->context_ID = kernel_record->context_ID;

    // Assign a mapping from this context ID to this process.
    context_ID_to_Process_map[process->context_ID] = process;

    // Assign the process an executable image by copying it from this
    // task's parent's process.
    Process* parent_process = task->parent_task->process;
    process->image_file_ID = parent_process->image_file_ID;
    strcpy(process->image_pathname, parent_process->image_pathname);

  } else {

    // A process already exists using the given context, so this task
    // is joining an existing process.  Grab that process.
    process = (*process_iterator).second;

    // Sanity check:  This process should explicitly indicate that it
    // is using the given context ID.
    if (process->context_ID != kernel_record->context_ID) {
      fprintf(stderr, "Consumer::handle_context_assignment(): ");
      fprintf(stderr, "Process and map disagree on context ID\n");
      exit(1);
    }

  }

  // Link the task and the process to one another.
  task->process = process;
  process->task_list.push_front(task);

} // Consumer::handle_context_assignment
// ===================================================================



// ===================================================================
void
Consumer::handle_duplicate_range () {

  // Get the virtual->canonical map for both the original (parent) and
  // duplicate (child) contexts.
  Process* original_process =
    safe_context_ID_to_process(kernel_record->context_ID);
  virtual_to_canonical_map_t* original_map =
    original_process->virtual_to_canonical_map;
  Process* duplicate_process =
    safe_context_ID_to_process(kernel_record->duplicate_context_ID);
  virtual_to_canonical_map_t* duplicate_map =
    duplicate_process->virtual_to_canonical_map;

  // Traverse every virtual page in the range, making that page in the
  // duplicate context point to the same canonical page as that page
  // does in the original context.
  for (virtual_page_ID_t current_page =
	 kernel_record->starting_virtual_page;
       current_page < kernel_record->ending_virtual_page;
       current_page++) {

    // We should never be trying to duplicate pages in the kernel's
    // address range.
    if (current_page >= KERNEL_VIRTUAL_PAGE_OFFSET) {
      fprintf(stderr, "Consumer::handle_duplicate_range(): ");
      fprintf(stderr, "In kernel range\n");
      exit(1);
    }

    // Attempt to lookup the canonical page underlying the original.
    // Note that it is possible for a task to map an anonymous
    // region without having referenced it yet, leaving some pages in
    // that region without assigned canonical pages.
    virtual_to_canonical_map_t::iterator i =
      (*original_map).find(current_page);

    // Has a canonical page yet been assigned to this page in the
    // original context?
    canonical_page_s* canonical_page = 0;
    if (i == (*original_map).end()) {

      // No.  File and shm pages are immediately assigned canonical
      // pages, so this must be an anonymous page.  Map it to the zero
      // page.
      map_virtual_to_canonical(original_map,
			       current_page,
			       canonical_zero_page);

    } else {

      // Yes.  Use that canonical page.
      canonical_page = (*i).second;

    }

    // The current page in the duplicate context cannot be mapped to
    // anything -- an error has occurred if it is.
    if ((*duplicate_map).find(current_page) !=
	(*duplicate_map).end()) {
      fprintf(stderr, "Consumer::handle_duplicate_range(): ");
      fprintf(stderr, "Duplicate range not unmapped\n");
      exit(1);
    }

    // Map this canonical page into the duplicate context.
    map_virtual_to_canonical(duplicate_map,
			     current_page,
			     canonical_page);

  }
} // Consumer::handle_duplicate_range
// ===================================================================



// ===================================================================
void
Consumer::do_file_mmap_range (const task_ID_t task_ID,
			      const file_ID_s& file_ID,
			      const virtual_page_ID_t starting_virtual_page,
			      const virtual_page_ID_t ending_virtual_page,
			      const virtual_page_ID_t starting_file_page,
			      const virtual_page_ID_t ending_file_page) {

  // Get the virtual->canonical maps for the context and the file.
  Task* task = safe_task_ID_to_task(task_ID);
  virtual_to_canonical_map_t* context_map =
    task->process->virtual_to_canonical_map;
  virtual_to_canonical_map_t* file_map =
    safe_file_ID_to_V2C_map(file_ID);

  // Traverse each page in the range, establishing an entry that maps
  // the task virtual page to the file page.
  virtual_page_ID_t task_page = starting_virtual_page;
  virtual_page_ID_t file_page = starting_file_page;
  while (task_page < ending_virtual_page) {

    // The task page should never be in the kernel's page range.
    if (task_page >= KERNEL_VIRTUAL_PAGE_OFFSET) {
      fprintf(stderr, "Consumer::do_file_mmap_range(): ");
      fprintf(stderr, "In kernel range\n");
      exit(1);
    }

    // Sanity check: We should never go off the end of the file page
    // range, either.
    assert(file_page < ending_file_page);

    // Attempt to lookup the canonical page underlying this file page.
    virtual_to_canonical_map_t::iterator i =
      (*file_map).find(file_page);

    // Has a canonical page yet been assigned to this file page?
    canonical_page_s* canonical_page = 0;
    if (i == (*file_map).end()) {

      // No.  Create a new canonical page and map it into the file's
      // space.
      canonical_page = get_new_canonical_page(TYPE_FILE);
      canonical_page->file_ID = file_ID;
      canonical_page->virtual_page = file_page;
      map_virtual_to_canonical(file_map, file_page, canonical_page);
      
    } else {

      // Yes.  Use that canonical page.
      canonical_page = (*i).second;

      // Is this page not a file page?
      if (canonical_page->type != TYPE_FILE) {

	// This is not a file cache page, so something has gone wrong
	// in mapping a canonical page that was of some other type.
	fprintf(stderr, "Consumer::do_file_mmap_range(): ");
	fprintf(stderr, "Mapping canonical page of wrong type\n");
	exit(1);

      }

      // Further sanity checks.
      if (!(canonical_page->file_ID == file_ID)) {
	assert(false);
      }
      assert(canonical_page->virtual_page == file_page);

    }

    // Ensure that the task doesn't already have a mapping for this
    // page in its space.  If so, emit the error and exit.
    if ((*context_map).find(task_page) != (*context_map).end()) {
      fprintf(stderr, "Consumer::do_file_mmap_range(): ");
      fprintf(stderr, "Attempt to map an already mapped page\n");
      exit(1);
    }

    // Map this canonical page into the task's space.
    map_virtual_to_canonical(context_map, task_page, canonical_page);
 
    // Move to the next page in each space.
    task_page++;
    file_page++;

  }

} // Consumer::do_file_mmap_range
// ===================================================================



// ===================================================================
void
Consumer::handle_file_mmap_range () {

  // Simply call on do_file_mmap_range() to do the real work.
  do_file_mmap_range(kernel_record->task_ID,
		     kernel_record->file_ID,
		     kernel_record->starting_virtual_page,
		     kernel_record->ending_virtual_page,
		     kernel_record->starting_file_page,
		     kernel_record->ending_file_page);

} // Consumer::handle_file_mmap_range
// ===================================================================



// ===================================================================
void
Consumer::do_anonymous_mmap_range
(const task_ID_t task_ID,
 const virtual_page_ID_t starting_virtual_page,
 const virtual_page_ID_t ending_virtual_page) {

  // Grab the virtual->canonical map for the process's context.
  Task* task = safe_task_ID_to_task(task_ID);
  virtual_to_canonical_map_t* map =
    task->process->virtual_to_canonical_map;

  // Traverse each page in the range, establishing an entry that maps
  // the task virtual page to the canonical ``zero page''.
  virtual_page_ID_t page = starting_virtual_page;
  while (page < ending_virtual_page) {

    // The task page should never be in the kernel's page range.
    if (page >= KERNEL_VIRTUAL_PAGE_OFFSET) {
      fprintf(stderr, "Consumer::do_anonymous_mmap_range(): ");
      fprintf(stderr,
	      "Page %lx in kernel range at record %qd\n",
	      page,
	      kernel_record->record_number);
      exit(1);
    }

    // Ensure that the task doesn't already have a mapping for this
    // page in its space.  If so, emit the error and exit.
    if ((*map).find(page) != (*map).end()) {
      fprintf(stderr, "Consumer::do_anonymous_mmap_range(): ");
      fprintf(stderr, "Attempt to map an already mapped page\n");
      exit(1);
    }

    // Map the canonical zero page into the task's space.  Note that
    // this mapping may not reflect the reality in the kernel, which
    // may not immediately map the virtual pages to any canonical
    // page.  However, it is safe to assume that it is mapped to the
    // zero page because any reads will require a mapping to that
    // page, and any writes will require a COW operation.  (See the
    // handling of reference records, where such COW operations are
    // performed, even in the absence of an explicit kernel COW
    // record.)
    map_virtual_to_canonical(map, page, canonical_zero_page);
 
    // Move to the next page.
    page++;

  }
} // Consumer::do_anonymous_mmap_range
// ===================================================================



// ===================================================================
void
Consumer::handle_anonymous_mmap_range () {

  // Simply call on do_anonymous_mmap_range() to do the real work.
  do_anonymous_mmap_range(kernel_record->task_ID,
			  kernel_record->starting_virtual_page,
			  kernel_record->ending_virtual_page);

} // Consumer::handle_anonymous_mmap_range
// ===================================================================



// ===================================================================
void
Consumer::do_shm_mmap_range (const task_ID_t task_ID,
			     const shm_ID_t shm_ID,
			     const virtual_page_ID_t starting_virtual_page,
			     const virtual_page_ID_t ending_virtual_page) {

  // Grab the virtual->canonical maps for the context and the shared
  // memory segment.
  Task* task = safe_task_ID_to_task(task_ID);
  virtual_to_canonical_map_t* context_map =
    task->process->virtual_to_canonical_map;
  virtual_to_canonical_map_t* shm_map =
    safe_shm_ID_to_V2C_map(shm_ID);

  // Traverse every page in the range, establishing an entry that maps
  // the task virtual page to the shared memory segment page.  Note
  // that shared memory segments are assumed to start at an offset of
  // zero.
  virtual_page_ID_t task_page = starting_virtual_page;
  virtual_page_ID_t shm_page = 0;
  while (task_page < ending_virtual_page) {

    // The task page should never be in the kernel's page range.
    if (task_page >= KERNEL_VIRTUAL_PAGE_OFFSET) {
      fprintf(stderr, "Consumer::do_shm_mmap_range(): In kernel range\n");
      exit(1);
    }

    // Attempt to lookup the canonical page underlying this shared
    // memory page.
    virtual_to_canonical_map_t::iterator i = (*shm_map).find(shm_page);

    // Has a canonical page yet been assigned to this shared memory
    // page?
    canonical_page_s* canonical_page = 0;
    if (i == (*shm_map).end()) {

      // No.  Create a new canonical page, associate it with this new
      // shm segment, and map it into the shared memory segment's
      // space, increasing the mapping count.
      canonical_page = get_new_canonical_page(TYPE_IPC_SHM_SEGMENT);
      canonical_page->shm_ID = shm_ID;
      canonical_page->virtual_page = shm_page;
      map_virtual_to_canonical(shm_map, shm_page, canonical_page);
      
    } else {

      // Yes.  Use that canonical page.
      canonical_page = (*i).second;

      // Sanity checks.
      assert(canonical_page->type == TYPE_IPC_SHM_SEGMENT);
      assert(canonical_page->shm_ID == shm_ID);
      assert(canonical_page->virtual_page == shm_page);

    }

    // Ensure that the task doesn't already have a mapping for this
    // page in its space.  If so, emit the error and exit.
    if ((*context_map).find(task_page) != (*context_map).end()) {
      fprintf(stderr, "Consumer::do_shm_mmap_range(): ");
      fprintf(stderr,
	      "Attempt to map an already mapped page %lx at record %qd\n",
	      task_page,
	      kernel_record->record_number);
      exit(1);
    }

    // Map this canonical page into the task's space.
    map_virtual_to_canonical(context_map, task_page, canonical_page);

    // Move to the next pair of pages.
    task_page++;
    shm_page++;

  }
} // Consumer::do_shm_mmap_range
// ===================================================================



// ===================================================================
void
Consumer::handle_shmat () {

  // Simply call on do_shm_mmap_range() to do the real work.
  do_shm_mmap_range(kernel_record->task_ID,
		    kernel_record->shm_ID,
		    kernel_record->starting_virtual_page,
		    kernel_record->ending_virtual_page);

} // Consumer::handle_shmat
// ===================================================================



// ===================================================================
void
Consumer::handle_shm_destroy () {

  // Grab the context for this shm segment.
  shm_ID_to_V2C_map_t::iterator map_iterator =
    shm_ID_to_V2C_map.find(kernel_record->shm_ID);

  // It may be the case that we've not encountered this shm segment
  // before, as it need not be attached to be destroyed by some
  // process.  There's only work to do if we've attached this shm
  // segment before.
  if (map_iterator != shm_ID_to_V2C_map.end()) {

    virtual_to_canonical_map_t* map =
      safe_shm_ID_to_V2C_map(kernel_record->shm_ID);

    // Clear the map contents, deleting its canonical pages.
    clear_address_space(map, TYPE_IPC_SHM_SEGMENT);

    // Delete the map and its association with the given shm ID.
    delete map;
    map = 0;
    shm_ID_to_V2C_map.erase(map_iterator);
    
  }

} // Consumer::handle_shm_destroy
// ===================================================================



// ===================================================================
// SFK: This method is horribly long and should, some day, be properly
// divided.
void
Consumer::handle_mremap () {

  // Grab the virtual->canonical map for the context.
  Task* task = safe_task_ID_to_task(kernel_record->task_ID);
  virtual_to_canonical_map_t* map =
    task->process->virtual_to_canonical_map;

  // Grab the starting virtual page numbers for the old and new
  // ranges
  virtual_page_ID_t old_page = kernel_record->old_starting_virtual_page;
  virtual_page_ID_t new_page = kernel_record->starting_virtual_page;

  // Sanity check: Since this is a Linux-specific system call, it
  // should be the case that the remapping covers a single VMA.
  // Therefore, the canonical page type and underlying address space
  // (e.g. file ID) should not change, and the offset within that
  // space (for files and shms) should increment regularly.  Copy the
  // starting canonical page use it to track progression through the
  // address space.
  virtual_to_canonical_map_t::iterator i = (*map).find(old_page);
  assert(i != (*map).end());
  canonical_page_s expected_page = *((*i).second);

  // Iterate through both ranges for the length of the old range.  For
  // each page visited in both ranges, map the new virtual page to the
  // same canonical page as the old virtual page.  Then unmap the old
  // virtual page.
  while (old_page < kernel_record->old_ending_virtual_page) {

    // Sanity check
    assert(new_page < kernel_record->ending_virtual_page);

    // Neither the old nor new virtual pages should be in the kernel's
    // page reange.
    if ((old_page >= KERNEL_VIRTUAL_PAGE_OFFSET) ||
	(new_page >= KERNEL_VIRTUAL_PAGE_OFFSET)) {
      fprintf(stderr, "Consumer::handle_mremap(): ");
      fprintf(stderr,
	      "old_page = %lx and new_page = %lx; in kernel range at record %qd\n",
	      old_page,
	      new_page,
	      kernel_record->record_number);
      exit(1);
    }

    // Ensure that the task doesn't already have a mapping for the new
    // page in its space.  If so, emit the error and exit.
    if ((*map).find(new_page) != (*map).end()) {
      fprintf(stderr, "Consumer::handle_mremap(): ");
      fprintf(stderr, "Attempt to map an already mapped page\n");
      exit(1);
    }

    // Grab the underlying canonical page for the old virtual page.
    // Sanity check: The page must exist, and it must match our
    // expected canonical page.  Note that ``matching'' can only be
    // fully tested for file and shm pages because they have a unique,
    // source context and virtual page ID within that context.
    i = (*map).find(old_page);
    assert(i != (*map).end());
    canonical_page_s* canonical_page = (*i).second;
    assert((expected_page.type == canonical_page->type) &&
	   ((expected_page.type == TYPE_ANONYMOUS) ||
	    ((expected_page.type == TYPE_FILE) &&
	     (expected_page.file_ID == canonical_page->file_ID) &&
	     (expected_page.virtual_page == canonical_page->virtual_page)) ||
	    ((expected_page.type == TYPE_IPC_SHM_SEGMENT) &&
	     (expected_page.shm_ID == canonical_page->shm_ID) &&
	     (expected_page.virtual_page == canonical_page->virtual_page))));

    // Create a mapping from the new virtual page to the canonical
    // page.
    map_virtual_to_canonical(map, new_page, canonical_page);

    // Remove the mapping from the old virtual page.
    (*map).erase(i);
    canonical_page->mapping_count--;

    // Move to the next old and new virtual pages, as well as the next
    // expected canonical page (for file and shm pages).
    old_page++;
    new_page++;
    expected_page.virtual_page++;

  }

  // Call on other functions to create the new mappings of additional
  // pages.
  switch (expected_page.type) {

  case TYPE_ANONYMOUS:

    do_anonymous_mmap_range(kernel_record->task_ID,
			    new_page,
			    kernel_record->ending_virtual_page);
    break;

  case TYPE_FILE:

    // The expected page is the next canonical page whose source is
    // the file associated with the mapping range being extended.
    // Thus, we can obtain the file ID and page range within that
    // file.  Note that the range of file pages can only be derived
    // from the range of virtual pages to be mapped.
    do_file_mmap_range(kernel_record->task_ID,
		       expected_page.file_ID,
		       new_page,
		       kernel_record->ending_virtual_page,
		       expected_page.virtual_page,
		       expected_page.virtual_page +
		       (kernel_record->ending_virtual_page - new_page));
    break;

  case TYPE_IPC_SHM_SEGMENT:

    // The expected page is the next canonical page whose source is
    // the shm segment associated with the mapping range being
    // extended.  Thus, We can obtain the shm ID from the expected
    // page.
    do_shm_mmap_range(kernel_record->task_ID,
		      expected_page.shm_ID,
		      expected_page.virtual_page,
		      kernel_record->ending_virtual_page);
    break;

  default:

    fprintf(stderr, "Consumer::handle_mremap(): ");
    fprintf(stderr,
	    "Canonical page with unknown type %c(%d) at record %qd\n",
	    expected_page.type,
	    (int)expected_page.type,
	    kernel_record->record_number);
    exit(1);

  }

} // Consumer::handle_mremap
// ===================================================================



// ===================================================================
void
Consumer::handle_munmap_range () {

  // Grab the virtual->canonical map for the context.
  Task* task = safe_task_ID_to_task(kernel_record->task_ID);
  virtual_to_canonical_map_t* map =
    task->process->virtual_to_canonical_map;

  // Traverse every page in the range, eliminating any mappings
  // found.
  virtual_page_ID_t page = kernel_record->starting_virtual_page;
  while (page < kernel_record->ending_virtual_page) {

    // The task page should never be in the kernel's page range.
    if (page >= KERNEL_VIRTUAL_PAGE_OFFSET) {
      fprintf(stderr, "Consumer::handle_munmap_range(): ");
      fprintf(stderr, "In kernel range\n");
      exit(1);
    }

    // Is this page mapped to another?
    virtual_to_canonical_map_t::iterator i = (*map).find(page);
    if (i != (*map).end()) {

      // It is, so eliminate the mapping.  If the underlying canonical
      // page is unused, delete if appropriate (determined within that
      // method).
      canonical_page_s* canonical_page = (*i).second;
      canonical_page->mapping_count--;
      (*map).erase(i);
      delete_if_unused_page(&canonical_page);

    }

    // Move to the next page.
    page++;

  }
} // Consumer::handle_munmap_range
// ===================================================================



// ===================================================================
void
Consumer::perform_complete_unmap (const context_ID_t context_ID) {

  // Is this context ID mapped to a process?  It may not be if the
  // process with which it was associated became defunct and was
  // deleted, taking the context with it.
  context_ID_to_Process_map_t::iterator process_iterator =
    context_ID_to_Process_map.find(context_ID);
  if (process_iterator == context_ID_to_Process_map.end()) {

    // The context has been deleted.  No work to do here.
    return;

  }

  // Grab the virtual->canonical map that forms the context for the
  // process.
  virtual_to_canonical_map_t* map =
    (*process_iterator).second->virtual_to_canonical_map;

  // This function takes a brute-force approach: It traverses every
  // page in the table, eliminating any mappings it might find.
  virtual_to_canonical_map_t::iterator i = (*map).begin();
  while (i != (*map).end()) {

    // It should never be the case that we are unmapping pages in the
    // kernel's page range.
    if ((*i).first >= KERNEL_VIRTUAL_PAGE_OFFSET) {
      fprintf(stderr, "Consumer::perform_complete_unmap(): ");
      fprintf(stderr, "In kernel range\n");
      exit(1);
    }

    // Eliminate the mapping.  Delete the underlying canonical page if
    // appropriate (handled within that method).
    canonical_page_s* canonical_page = (*i).second;
    virtual_to_canonical_map_t::iterator moribund = i;
    i++;
    (*map).erase(moribund);
    canonical_page->mapping_count--;
    delete_if_unused_page(&canonical_page);

  }
} // Consumer::perform_complete_unmap
// ===================================================================



// ===================================================================
void
Consumer::handle_cow_unmap () {

  virtual_page_ID_t page = kernel_record->starting_virtual_page;

  // The page should not be in the kernel's page range.
  if (page >= KERNEL_VIRTUAL_PAGE_OFFSET) {
    fprintf(stderr, "Consumer::handle_cow_unmap(): In kernel range\n");
    exit(1);
  }

  // Grab the virtual->canonical map for the context.
  Task* task = safe_task_ID_to_task(kernel_record->task_ID);
  virtual_to_canonical_map_t* map =
    task->process->virtual_to_canonical_map;

  // This page should be mapped.  If it is not, an error has
  // occurred.
  virtual_to_canonical_map_t::iterator i = (*map).find(page);
  if (i == (*map).end()) {
    fprintf(stderr, "Consumer::handle_cow_unmap(): ");
    fprintf(stderr, "Unmapping a page with no mapping ");
    fprintf(stderr, "at record %qd\n", kernel_record->record_number);
    exit(1);
  }

  // The original page cannot have been an IPC shared memory page.
  canonical_page_s* original_canonical_page = (*i).second;
  if (original_canonical_page->type == TYPE_IPC_SHM_SEGMENT) {
    fprintf(stderr, "Consumer::handle_cow_unmap(): ");
    fprintf(stderr, "COW with an IPC shm page\n");
    exit(1);
  }

  // When COW-unmapping an anonymous page, it should never be the case
  // that the sharing count reaches zero -- the last task to write
  // to the COW page should get that original.
  //
  // C: We _do_ see this phenomenon.  Why?  Harmless (we think) to let
  // it go...
  if ((original_canonical_page->type == TYPE_ANONYMOUS) &&
      (original_canonical_page->mapping_count == 1)) {
    fprintf(stderr, "WARNING: Consumer::handle_cow_unmap(): ");
    fprintf(stderr,
	    "Anonymous page orphaned at record %qd\n",
	    kernel_record->record_number);
    // exit(1);
  }

  // Reassign this virtual page to a new canonical page, thus
  // representing the copy that is made by the COW mechanism.  Whether
  // the original page was anonymous or from a file, the new copy must
  // be anonymous.  Also increase its sharing count.
  original_canonical_page->mapping_count--;
  (*map).erase(i);
  delete_if_unused_page(&original_canonical_page);
  canonical_page_s* new_canonical_page =
    get_new_canonical_page(TYPE_ANONYMOUS);
  map_virtual_to_canonical(map, page, new_canonical_page);

} // Consumer::handle_cow_unmap
// ===================================================================



// ===================================================================
void
Consumer::handle_file_read_or_write () {

  // Get the virtual->canonical map for the file.
  virtual_to_canonical_map_t* map =
    safe_file_ID_to_V2C_map(kernel_record->file_ID);

  // Traverse each page in the requested read or write operation.
  virtual_page_ID_t page = kernel_record->starting_file_page;
  while (page < kernel_record->ending_file_page) {

    // Attempt to lookup the canonical page underlying this file page.
    virtual_to_canonical_map_t::iterator i = (*map).find(page);

    // Has a canonical page yet been assigned to this file page?
    canonical_page_s* canonical_page = 0;
    if (i == (*map).end()) {

      // No.  Create a new canonical page and map it into the file's
      // space.
      canonical_page = get_new_canonical_page(TYPE_FILE);
      canonical_page->file_ID = kernel_record->file_ID;
      canonical_page->virtual_page = page;      
      map_virtual_to_canonical(map, page, canonical_page);

    } else {

      // Yes.  Use that canonical page.
      canonical_page = (*i).second;

      // Is this page not a file page?
      if (canonical_page->type != TYPE_FILE) {

	// This is not a file cache page, so something has gone wrong
	// in mapping a canonical page that was of some other type.
	fprintf(stderr, "Consumer::handle_file_read_or_write(): ");
	fprintf(stderr, "Found canonical page of the wrong type\n");
	exit(1);

      }

      // Further sanity checks.
      assert(canonical_page->file_ID == kernel_record->file_ID);
      assert(canonical_page->virtual_page == page);

    }

    // Advance to the next page of the file.
    page++;

  }

} // Consumer::handle_file_read_or_write
// ===================================================================



// ===================================================================
void
Consumer::handle_file_delete () {

  // Grab the context for this file.
  file_ID_to_V2C_map_t::iterator map_iterator =
    file_ID_to_V2C_map.find(kernel_record->file_ID);

  // It may be the case that we've not encountered this file before,
  // as a file need not be mapped or opened to be deleted by some
  // process.  There's only work to do if we've a record of this
  // file.
  if (map_iterator != file_ID_to_V2C_map.end()) {

    virtual_to_canonical_map_t* map =
      safe_file_ID_to_V2C_map(kernel_record->file_ID);

    // Clear the map contents, deleting its canonical pages.
    clear_address_space(map, TYPE_FILE);

    // Delete the map and its association with the given file ID.
    delete map;
    map = 0;
    file_ID_to_V2C_map.erase(map_iterator);

  } // File previously encountered

} // Consumer::handle_file_delete
// ===================================================================



// ===================================================================
void
Consumer::handle_invalid () {

  fprintf(stderr,
	  "Invalid kernel event tag %c(%d) at record %qd\n",
	  kernel_record->tag,
	  (int)kernel_record->tag,
	  kernel_record->record_number);
  exit(1);

} // Consumer::handle_invalid
// ===================================================================



// ===================================================================
void
Consumer::update_with_reference_record () {

  // If this is the first observation of a reference, mark the change
  // of phases to waiting for a scheduling event.
  if (merge_phase == PHASE_PRE_REFERENCE) {
    merge_phase = PHASE_WAITING_FOR_SCHEDULE;
  }

  // If we're still waiting to see the first scheduling event (which
  // would tell us which task is active), then we cannot yet process
  // this record.
  if (merge_phase == PHASE_WAITING_FOR_SCHEDULE) {
    return;
  }

  // Grab local copies of the referenced page's information.
  const context_ID_t context_ID =
    reference_record->virtual_page.context_ID;
  const virtual_page_ID_t page_ID =
    reference_record->virtual_page.page_ID;

  // Sanity check: The context used should be the scheduled task's.
  if (context_ID != scheduled_task->process->context_ID) {
    fprintf(stderr, "Consumer::update_with_reference_record(): ");
    fprintf(stderr,
	    "Unexpected context ID used at record %qd\n",
	    reference_record->record_number);
    exit(1);
  }

  // Sanity check: The referenced page cannot be in the kernel's
  // space.
  if (page_ID >= KERNEL_VIRTUAL_PAGE_OFFSET) {
    fprintf(stderr, "Consumer::update_with_reference_record(): ");
    fprintf(stderr,
	    "Invalid page ID %lx (in kernel space) at record %qd\n",
	    page_ID,
	    reference_record->record_number);
    exit(1);
  }

  // Grab the map of the context in which this reference was
  // performed.
  Process* process = safe_context_ID_to_process(context_ID);
  virtual_to_canonical_map_t* map = process->virtual_to_canonical_map;

  // Sanity check: The reference must be to some mapped page.  Since
  // all mappings are recorded in the kernel trace including brk()
  // calls, no references should exist to unmapped pages.
  virtual_to_canonical_map_t::iterator i = (*map).find(page_ID);
  if (i == (*map).end()) {
    fprintf(stderr, "Consumer::update_with_reference_record(): ");
    fprintf(stderr,
	    "Invalid page ID %lx (unmapped region) for task ID %x at record %qd\n",
	    page_ID,
	    scheduled_task->task_ID,
	    reference_record->record_number);
    fprintf(stderr,
	    "cycle time = %qx\n",
	    reference_record->cycle_timestamp);
    exit(1);
  }

  // If this is a write operation to the zero page, then we must
  // create a new canonical page.  Note that in the real OS, the page
  // must initially not have ever been mapped to the zero page -- we
  // just assumed it as a default for the anonymous pages.  This
  // operation is effectively performing a COW from a canonical page
  // that the original system never mapped.
  canonical_page_s* canonical_page = (*i).second;
  if (((reference_record->tag == TAG_SINGLE_WRITE) ||
       (reference_record->tag == TAG_MULTIPLE_WRITE)) &&
      (canonical_page == canonical_zero_page)) {

    // Unmap the canonical page.
    canonical_page->mapping_count--;
    map->erase(i);
    delete_if_unused_page(&canonical_page);

    // Create a new anonymous, canonical page and map to it.
    canonical_page = get_new_canonical_page(TYPE_ANONYMOUS);
    map_virtual_to_canonical(map, page_ID, canonical_page);

  }

} // Consumer::update_with_reference_record
// ===================================================================



// ===================================================================
void
Consumer::update_with_kernel_record () {

  handler_t handler =
    handler_dispatch_table[(int)kernel_record->tag];
  if (handler == (handler_t)NULL) {

    fprintf(stderr, "Consumer::update_with_kernel_record(): ");
    fprintf(stderr,
	    "At record %qd: Unknown tag %c(%d)\n",
	    kernel_record->record_number,
	    kernel_record->tag,
	    (int)kernel_record->tag);
    exit(1);

  }

  ((*this).*(handler))();

} // Consumer::update_with_kernel_record
// ===================================================================



// ===================================================================
canonical_page_ID_t
Consumer::lookup_canonical_page_ID
(const virtual_to_canonical_map_t* const map,
 const virtual_page_ID_t virtual_page_ID) {

  canonical_page_s* canonical_page =
    safe_virtual_to_canonical_page(map, virtual_page_ID);

  return canonical_page->page;

} // lookup_canonical_page_ID
// ===================================================================



// ===================================================================
void
Consumer::initialize () {

  // Create a task object for the initial kernel task as well as a
  // process object for the ``kernel process''.
  kernel_task = allocate_new_task();
  Process* kernel_process = allocate_new_process();

  // Give each an ID of 0, initialize the other fields, and assign the
  // thread to the process.
  kernel_task->task_ID = 0;
  kernel_task->process = kernel_process;
  kernel_task->parent_task = NULL;
  kernel_process->process_ID = KERNEL_PROCESS_ID;
  kernel_process->task_list.push_front(kernel_task);
  kernel_process->image_file_ID.inode_ID = 0;
  kernel_process->image_file_ID.major = 0;
  kernel_process->image_file_ID.minor = 0;
  strcpy(kernel_process->image_pathname, "***KERNEL***");

  // Kernel threads may create mappings into the kernel's virtual
  // address space.  We won't see any references to those, but we'll
  // process them mappings.
  kernel_process->context_ID = KERNEL_CONTEXT;
  kernel_process->virtual_to_canonical_map =
    new virtual_to_canonical_map_t;
  if (kernel_process->virtual_to_canonical_map == NULL) {
    fprintf(stderr, "Consumer::Consumer(): ");
    fprintf(stderr, "Failed allocation of kernel virtual->canonical map\n");
    exit(1);
  }

  // Insert into both the task_ID- and context_ID-keyed tables.
  task_ID_to_Task_map[kernel_task->task_ID] = kernel_task;
  context_ID_to_Process_map[kernel_process->context_ID] =
    kernel_process;

  // Assume that the base kernel task is the first one scheduled.
  scheduled_task = kernel_task;

}
// ===================================================================



// ===================================================================
Task*
Consumer::allocate_new_task () {

  // Attempt to allocate a new Task.  Check for success.
  Task* task = new Task;
  if (task == NULL) {
    fprintf(stderr, "Consumer::allocate_new_task(): ");
    fprintf(stderr, "Allocation failure\n");
    exit(1);
  }

  return task;

} // Consumer::allocate_new_task
// ===================================================================



// ===================================================================
Process*
Consumer::allocate_new_process () {

  // Attempt to allocate a new Task.  Check for success.
  Process* process = new Process;
  if (process == NULL) {
    fprintf(stderr, "Consumer::allocate_new_process(): ");
    fprintf(stderr, "Allocation failure\n");
    exit(1);
  }

  return process;

} // Consumer::allocate_new_process
// ===================================================================
