/**
 * \file Per_Process_Consumer.hh
 * \author Scott F. H. Kaplan -- sfkaplan@cs.amherst.edu
 * \date September 2004
 * \brief Produces per-process reference files.
 *
 * A subclass of Consumer that divides the stream of reference and
 * kernel records on a per-task basis, and groups tasks into
 * processes.  One output file is created for each process, where that
 * file will contain every reference performed by the tasks of that
 * process as well as every system call (e.g. blocking I/O or file
 * operations).  The original scheduling of those tasks is reflected
 * in each per-process trace.
 *
 * Virtual memory references emitted will have the canonical page ID
 * substituted for the virtual page number and context IDs.
 * Similarly, file system operations will be substituted with a
 * sequence of references to canonical page IDs.
 **/



// ===================================================================
// Avoid multiple inclusion.
#if !defined (_PER_PROCESS_CONSUMER_HH)
#define _PER_PROCESS_CONSUMER_HH
// ===================================================================



// ===================================================================
// INCLUDES

#include "Consumer.hh"
#include "Per_Process_Process.hh"
#include "Task.hh"
// ===================================================================



// ===================================================================
// CLASS DEFINITION

/**
 * \brief Divide reference streams by process and task.
 *
 * Produce multiple traces grouped by process and, within each
 * process, by thread (a.k.a. task).
 **/
class Per_Process_Consumer : public Consumer {

public: // CONSTRUCTOR

  /**
   * \brief The constructor.
   * \param reader The object that will read and parse the reference
   *        and kernel event traces.
   * \param output_directory_pathname The pathname of the directory
   *        that will hold all of the output files.
   *
   * Store the output directory pathname for later, when processes are
   * encountered and require new files to be created.
   **/
  Per_Process_Consumer (Reader* const reader,
			const char* const output_directory_pathname);

  /**
   * \brief The deconstructor.
   *
   * Close all per-process writers.
   **/
  virtual ~Per_Process_Consumer ();

protected: // VIRTUAL METHODS

  /** \brief Called after processing ends -- Does nothing. */
  virtual void clean_up () {}

  /**
   * \brief Called after processing a reference record.
   *
   * Direct the reference to its process's writer, emitting the
   * reference with the canonical page number.
   **/
  virtual void act_on_reference_record ();

  /** \brief Called after processing a kernel record.
   *
   * If it is a system call performed by the process, emit it to the
   * process's writer.  If the record indicates a fork, exit, or
   * context assignment, update the per-process writers and mappings.
   **/
  virtual void act_on_kernel_record ();

  /**
   * \brief Allocate a new per-process process object.
   * \return A pointer to a newly-allocated process object.
   *
   * Allocate a new process object, where the object is a
   * Per_Process_Process object, designed to additionally associate a
   * process with its writer.  Abort if the allocation fails.
   **/
  virtual Process* allocate_new_process ();

protected: // SUPPORT METHODS

  /**
   * \brief Create a new writer for a given process.
   * \param process The per-process process for which to create a
   *        writer.
   *
   * For a per-process process that has no writer, create one using
   * the process ID.
   **/
  void create_writer (Per_Process_Process* process);

protected: // DATA MEMBERS

  /** \brief The directory into which all results will be placed. */
  char* output_directory_pathname;

}; // class Per_Process_Consumer
// ===================================================================



// ===================================================================
#endif // _PER_PROCESS_CONSUMER_HH
// ===================================================================
