/**
 * \file Per_Process_Writer.hh
 * \author Scott F. H. Kaplan <sfkaplan@cs.amherst.edu>
 * \date June 2004
 * \brief The header for the Per_Process_Writer class.
 *
 * The definition of a Per_Process_Writer, designed to emit the
 * references and kernel events of a single process, where the timing
 * information is virtualized to reflect only this process.
 **/



// =================================================================
#if !defined (_PER_PROCESS_WRITER_HH)
#define _PER_PROCESS_WRITER_HH
// =================================================================



// =================================================================
// INCLUDES

#include <zlib.h>
#include "types-and-constants.hh"
// =================================================================



// =================================================================
/**
 * \brief A Per_Process_Writer emits references and kernel events for
 *        a single process.
 *
 * Accept a stream of reference and kernel events, and emit them to a
 * file associated with a single process.  Keep cycle, instruction,
 * and reference clocks up-to-date for the one process in isolation,
 * thus emitting traces that provide timestamps that a virtualized to
 * that process.  Since the process may reflect many threads, we log
 * the original scheduling of those threads for this process.
 **/
class Per_Process_Writer {

public:

  /**
   * \brief The constructor.
   * \see append_reference
   * \see append_kernel_event
   * \param output_directory_pathname The pathname for the directory
   *        of output traces.
   * \param process_ID The ID of the process begin logged.
   *
   * Prepare a writer by constructing the pathname within the base
   * directory for the given process ID.  The file will not be opened
   * until at least one record (reference or kernel) is appended to
   * the output stream.  All reference and kernel records submitted to
   * this object will be emitted into this file.
   **/
  Per_Process_Writer (const char* const output_directory_pathname,
		      const process_ID_t process_ID);

  /**
   * \brief The destructor.
   *
   * Flush and close the output file.
   **/
  ~Per_Process_Writer ();

  /**
   * \brief Emit the next reference record.
   * \param tag The tag indicating the type of reference to emit.
   * \param canonical_page_ID The canonical number of the page that is
   *        referenced.
   * \param virtual_cycle_time The current virtual cycle time of the
   *        process performing the reference.
   * \param virtual_reference_time The current virtual reference time
   *        of the process performing the reference.
   *
   * Emit this record into the output stream.  If the output stream
   * has not actually been opened (that is, this is the first record
   * submitted to this stream), then open the output file.
   **/
  void append_reference_record
  (const tag_t tag,
   const canonical_page_ID_t canonical_page_ID,
   const timestamp_t virtual_cycle_time,
   const timestamp_t virtual_reference_time);

  /**
   * \brief Emit the next kernel record.
   * \param kernel_record The record whose contents is to be emitted.
   * \param virtual_cycle_time The current virtual cycle time of the
   *        process performing the reference.
   * \param virtual_reference_time The current virtual reference time
   *        of the process performing the reference.
   *
   * Emit this record into the output stream.  If the output stream
   * has not actually been opened (that is, this is the first record
   * submitted to this stream), then open the output file.
   **/
  void append_kernel_record
  (const kernel_record_s* const kernel_record,
   const timestamp_t virtual_cycle_time,
   const timestamp_t virtual_reference_time);

protected: // METHODS

  /**
   * \brief Open the compressed output file.
   *
   * Open the output file, this providing a compressed stream to which
   * records can be written.  Note that this method should be called
   * only once, upon the first addition of a record to the output
   * stream.
   **/
  void open_output ();

protected: // DATA MEMBERS

  /** \brief The directory of the compressed output stream. */
  char* output_directory;

  /** \brief The full pathname of the compressed output stream. */
  char* output_pathname;

  /** \brief The compressed output stream to which reduced records are
   *         written. */
  gzFile output_stream;

  /** \brief The cycle timestamp of the previously emitted record. */
  timestamp_t previous_cycle_time;

  /** \brief The reference timestamp of the previously emitted
   *         record. */
  timestamp_t previous_reference_time;

}; // Per_Process_Writer
// =================================================================



// =================================================================
#endif // _PER_PROCESS_WRITER_HH
// =================================================================
