/**
 * \file Per_Task_Consumer.cc
 * \author Scott F. H. Kaplan -- sfkaplan@cs.amherst.edu
 * \date June 2004
 * \brief Produces per-task reference files, grouped by process.
 **/



// ===================================================================
// INCLUDES

#include <assert.h>
#include <stdio.h>
#include <string.h>
#include "Per_Task_Consumer.hh"
// ===================================================================



// ===================================================================
Per_Task_Consumer::Per_Task_Consumer
(Reader* const reader,
 const char* const output_directory_pathname) : Consumer(reader) {

   // Allocate space for the output directory pathname.
   this->output_directory_pathname = new char[BUFFER_SIZE];
   if (output_directory_pathname == NULL) {
     fprintf(stderr, "Per_Task_Consumer::Per_Task_Consumer(): ");
     fprintf(stderr, "Could not allocate space for output directory\n");
     exit(1);
  }

  // Copy the output directory pathname into this object.
  strcpy(this->output_directory_pathname, output_directory_pathname);

} // Per_Task_Consumer::Per_Task_Consumer
// ===================================================================



// ===================================================================
Per_Task_Consumer::~Per_Task_Consumer () {

  // Traverse the task_ID->Task table.
  task_ID_to_Task_map_t::iterator i = task_ID_to_Task_map.begin();
  while (i != task_ID_to_Task_map.end()) {

    // Delete the per-task writer associated with this task, thus
    // flushing and closing its output.
    Per_Task_Task* task = (Per_Task_Task*)(*i).second;
    delete task->writer;
    task->writer = NULL;

    // Move to the next task.
    i++;

  } // task_ID->Task table traversal

} // Per_Task_Consumer::~Per_Task_Consumer
// ===================================================================



// ===================================================================
void
Per_Task_Consumer::act_on_reference_record () {

  // If we're still waiting to see the first scheduling event (which
  // would tell us which task is active), then we cannot yet process
  // this record.
  assert(merge_phase != PHASE_PRE_REFERENCE);
  if (merge_phase != PHASE_PROCESSING_REFERENCES) {
    return;
  }

  // Lookup the canonical page ID for this reference in the scheduled
  // task's context.
  canonical_page_ID_t canonical_page_ID =
    lookup_canonical_page_ID(scheduled_task->process->virtual_to_canonical_map,
			     reference_record->virtual_page.page_ID);

  // Submit this reference to this task's writer.
  Per_Task_Task* scheduled_ptt = (Per_Task_Task*)scheduled_task;
  if (scheduled_ptt->writer == NULL) {
    create_writer(scheduled_ptt);
  }
  scheduled_ptt->writer->
    append_reference_record(reference_record->tag,
			    canonical_page_ID,
			    scheduled_task->virtual_cycle_time,
			    scheduled_task->virtual_reference_time);

} // Consumer::act_on_reference_record
// ===================================================================



// ===================================================================
void
Per_Task_Consumer::act_on_kernel_record () {

  // Determine how to react to this record depending on the type of
  // kernel event.
  switch (kernel_record->tag) {

  case TAG_CONTEXT_ASSIGNMENT:

    act_on_context_assignment();
    break;

  case TAG_FILE_OPEN:
  case TAG_FILE_CLOSE:
  case TAG_ACCEPT: {

    // These kernel events are potentially blocking, and thus should
    // appear in the output trace for the current task.  However, we
    // only care about them if reference processing is occuring.
    if (merge_phase != PHASE_PROCESSING_REFERENCES) {
      return;
    }

    // As sanity check, ensure that the task ID given for this record
    // matches the current task's ID.
    if (kernel_record->task_ID != scheduled_task->task_ID) {
      fprintf(stderr, "Per_Task_Consumer::act_on_kernel_record(): ");
      fprintf(stderr, "Record and scheduled task ID mismatch\n");
      exit(1);
    }

    // Pass the kernel event to the task's writer.  Ensure first that
    // the writer exists.
    Per_Task_Task* scheduled_ptt = (Per_Task_Task*)scheduled_task;
    if (scheduled_ptt->writer == NULL) {
      create_writer(scheduled_ptt);
    }
    scheduled_ptt->writer->
      append_kernel_record(kernel_record,
			   scheduled_task->virtual_cycle_time,
			   scheduled_task->virtual_reference_time);

    break;

  }

  case TAG_FILE_READ:
  case TAG_FILE_WRITE: {

    // These kernel events are really referencing events, so let's
    // treat them as such.  Lookup the virtual->canonical map for the
    // file.
    virtual_to_canonical_map_t* map =
      safe_file_ID_to_V2C_map(kernel_record->file_ID);

    // Traverse each page in the requested read or write operation.
    virtual_page_ID_t page = kernel_record->starting_file_page;
    while (page < kernel_record->ending_file_page) {

      // Lookup the canonical page underlying this file page.
      virtual_to_canonical_map_t::iterator i = (*map).find(page);
      assert(i != (*map).end());
      canonical_page_ID_t canonical_page_ID = (*i).second->page;

      // Submit this reference to this task's writer.
      Per_Task_Task* scheduled_ptt = (Per_Task_Task*)scheduled_task;
      if (scheduled_ptt->writer == NULL) {
	create_writer(scheduled_ptt);
      }
      scheduled_ptt->writer->
	append_reference_record(kernel_record->tag,
				canonical_page_ID,
				scheduled_task->virtual_cycle_time,
				scheduled_task->virtual_reference_time);

      // Advance to the next page of the file.
      page++;

    }

    break;

  }

  default:

    // Do nothing -- ignore this kernel record, as it's not relevant
    // to the output of these per-task traces.
    break;

  }

} // Consumer::act_on_kernel_record
// ===================================================================



// ===================================================================
Task*
Per_Task_Consumer::allocate_new_task () {

  // Attempt to allocate a Per_Task_Task, a specific subclass designed for
  // use with this class.
  Per_Task_Task* task = new Per_Task_Task;
  if (task == NULL) {
    fprintf(stderr, "Per_Task_Consumer::allocate_new_task(): ");
    fprintf(stderr, "Allocation failure\n");
    exit(1);
  }

  return task;

} // Per_Task_Consumer::allocate_new_task
// ===================================================================



// ===================================================================
void
Per_Task_Consumer::create_writer (Per_Task_Task* task) {

  // Create a new writer for this task and its newly assigned context
  // and process.
  task->writer = new Per_Task_Writer(output_directory_pathname,
				     task->process->process_ID,
				     task->task_ID);
  if (task->writer == NULL) {
    fprintf(stderr, "Per_Task_Consumer::create_writer(): ");
    fprintf(stderr, "Failed to allocate new per-task writer\n");
    exit(1);
  }

} // Per_Task_Consumer::create_writer
// ===================================================================



// ===================================================================
void
Per_Task_Consumer::act_on_context_assignment () {

  // Look up the task being assigned a context.
  Per_Task_Task* task =
    (Per_Task_Task*)safe_task_ID_to_task(kernel_record->task_ID);

  // Delete (and thus flush and close) the old writer if it exists.  A
  // new writer will be created at the next output operation.
  if (task->writer == NULL) {
    delete task->writer;
    task->writer = NULL;
  }

} // Per_Task_Consumer::act_on_context_assignment
// ===================================================================
