/**
 * \file Per_Task_Consumer.hh
 * \author Scott F. H. Kaplan -- sfkaplan@cs.amherst.edu
 * \date June 2004
 * \brief Produces per-task reference files, grouped by process.
 *
 * A subclass of Consumer that divides the stream of reference and
 * kernel records on a per-task basis, and groups tasks into
 * processes.  One output file is created for each task, where that
 * file will contain every reference performed by that task as well as
 * every system call (e.g. blocking I/O or file operations).  For each
 * process (that is, group of 1 or more tasks sharing a virtual
 * address space), a separate directory is created and the output
 * files for its tasks are placed in that directory.
 *
 * Virtual memory references emitted will have the canonical page ID
 * substituted for the virtual page number and context IDs.
 * Similarly, file system operations will be substituted with a
 * sequence of references to canonical page IDs.
 **/



// ===================================================================
// Avoid multiple inclusion.
#if !defined (_PER_TASK_CONSUMER_HH)
#define _PER_TASK_CONSUMER_HH
// ===================================================================



// ===================================================================
// INCLUDES

#include "Consumer.hh"
#include "Per_Task_Task.hh"
// ===================================================================



// ===================================================================
// CLASS DEFINITION

/**
 * \brief Divide reference streams by process and task.
 *
 * Produce multiple traces grouped by process and, within each
 * process, by thread (a.k.a. task).
 **/
class Per_Task_Consumer : public Consumer {

public: // CONSTRUCTOR

  /**
   * \brief The constructor.
   * \param reader The object that will read and parse the reference
   *        and kernel traces.
   * \param output_directory_pathname The pathname of the directory
   *        that will hold all of the output files.
   *
   * Store the output directory pathname for later, when tasks and
   * processes are encountered and require new directories and files
   * to be created.
   **/
  Per_Task_Consumer (Reader* const reader,
		     const char* const output_directory_pathname);

  /**
   * \brief The deconstructor.
   *
   * Close all per-task writers.
   **/
  virtual ~Per_Task_Consumer ();

protected: // VIRTUAL METHODS

  /** \brief Called after processing ends -- Does nothing. */
  virtual void clean_up () {}

  /**
   * \brief Called after processing a reference record.
   *
   * Direct the reference to its task's writer, emitting the reference
   * with the canonical page number.
   **/
  virtual void act_on_reference_record ();

  /** \brief Called after processing a kernel record.
   *
   * If it is a system call performed by the task, emit it to the
   * task's writer.  If the record indicates a fork, exit, or context
   * assignment, update the per-task writers and mappings.
   **/
  virtual void act_on_kernel_record ();

  /**
   * \brief Allocate a new per-task task object.
   * \return A pointer to a newly-allocated task object.
   *
   * Allocate a new task object, where the object is a Per_Task_Task
   * object, designed to additionally associate a task with its
   * writer.  Abort if the allocation fails.
   **/
  virtual Task* allocate_new_task ();

protected: // SUPPORT METHODS

  /**
   * \brief Create a new writer for a given task.
   * \param task The per-task task for which to create a writer.
   *
   * For a per-task task that has no writer, create one using the
   * task's process and task IDs.
   **/
  void create_writer (Per_Task_Task* task);

  /**
   * \brief Close the current writer in response to a context change.
   *
   * If the new context associates the task with a different process,
   * close the current writer.
   **/
  void act_on_context_assignment ();

protected: // DATA MEMBERS

  /** \brief The directory into which all results will be placed. */
  char* output_directory_pathname;

}; // class Per_Task_Consumer
// ===================================================================



// ===================================================================
#endif // _PER_TASK_CONSUMER_HH
// ===================================================================
