/**
 * \file Per_Task_Writer.cc
 * \author Scott F. H. Kaplan <sfkaplan@cs.amherst.edu>
 * \date June 2004
 * \brief The code for the Per_Task_Writer class.
 *
 * The methods of a Per_Task_Writer, which is responsible for emitting
 * records into a trace of reference and kernel events for one task.
 **/



// ===================================================================
// INCLUDES

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include "Per_Task_Writer.hh"
// ===================================================================



// ===================================================================
Per_Task_Writer::Per_Task_Writer
(const char* const output_directory_pathname,
 const process_ID_t process_ID,
 const task_ID_t task_ID) {

  output_directory = new char[BUFFER_SIZE];
  output_pathname = new char[BUFFER_SIZE];
  if ((output_directory == NULL) || (output_pathname == NULL)) {
    fprintf(stderr, "Per_Task_Writer::Per_Task_Writer(): ");
    fprintf(stderr, "Failed allocation of output directory or pathname\n");
    exit(1);
  }

  // Construct the output directory for later use and nullify the
  // output stream.  (The output file is opened only when the first
  // record is added to the filter.)
  sprintf(output_directory,
	  "%s/%x",
	  output_directory_pathname,
	  process_ID);
  sprintf(output_pathname,
	  "%s/%hx.gz",
	  output_directory,
	  task_ID);
  output_stream = NULL;

  // Initialize the ``last record'' values to zero, which is the
  // assumed starting point for difference encoding purposes.
  previous_cycle_time = 0;
  previous_reference_time = 0;

} // Per_Task_Writer::Per_Task_Writer
// ===================================================================



// ===================================================================
Per_Task_Writer::~Per_Task_Writer () {

  // There is only work to do if the output stream was opened.
  if (output_stream != NULL) {

    // Close the output stream, ensuring that is succeeded.
    int result = gzclose(output_stream);
    if (result != Z_OK) {
      fprintf(stderr, "Per_Task_Writer::~Per_Task_Writer(): ");
      fprintf(stderr, "Close failure\n");
      exit(1);
    }
  }

} // Per_Task_Writer::~Per_Task_Writer
// ===================================================================



// ===================================================================
void
Per_Task_Writer::append_reference_record
(const tag_t tag,
 const canonical_page_ID_t canonical_page_ID,
 const timestamp_t virtual_cycle_time,
 const timestamp_t virtual_reference_time) {

  // If the output stream is not yet open, then open it.
  if (output_stream == NULL) {
    open_output();
  }

  // Calculate the differences for each timestamp (always positive).
  timestamp_t cycle_timestamp_delta =
    virtual_cycle_time - previous_cycle_time;
  timestamp_t reference_timestamp_delta =
    virtual_reference_time - previous_reference_time;

  // Store the current timestamps for later.
  previous_cycle_time = virtual_cycle_time;
  previous_reference_time = virtual_reference_time;

  // Print the difference encoded values to the compressed stream.
  int result = gzprintf(output_stream,
			"%c %qx %qx %lx\n",
			tag,
			cycle_timestamp_delta,
			reference_timestamp_delta,
			canonical_page_ID);

  // Ensure that the print succeeded.
  if (result <= 0) {
    fprintf(stderr, "Per_Task_Writer::append_reference_record(): ");
    fprintf(stderr, "Print failed\n");
    exit(1);
  }

} // Per_Task_Writer::append_reference_record
// ===================================================================



// ===================================================================
void
Per_Task_Writer::append_kernel_record
(const kernel_record_s* const kernel_record,
 const timestamp_t virtual_cycle_time,
 const timestamp_t virtual_reference_time) {

  // If the output stream is not yet open, then open it.
  if (output_stream == NULL) {
    open_output();
  }

  // Calculate the differences for each timestamp (always positive).
  timestamp_t cycle_timestamp_delta =
    virtual_cycle_time - previous_cycle_time;
  timestamp_t reference_timestamp_delta =
    virtual_reference_time - previous_reference_time;

  // Store the current timestamps for later.
  previous_cycle_time = virtual_cycle_time;
  previous_reference_time = virtual_reference_time;

  // Determine how to emit this record based on its type.
  int result = 0;
  switch(kernel_record->tag) {

  case TAG_FILE_OPEN:

    result = gzprintf(output_stream,
		      "%c %qx %qx %lx %hd %hd %s\n",
		      kernel_record->tag,
		      cycle_timestamp_delta,
		      reference_timestamp_delta,
		      kernel_record->file_ID.inode_ID,
		      kernel_record->file_ID.major,
		      kernel_record->file_ID.minor,
		      kernel_record->filename);
    break;

  case TAG_FILE_CLOSE:

    result = gzprintf(output_stream,
		      "%c %qx %qx %lx %hd %hd\n",
		      kernel_record->tag,
		      cycle_timestamp_delta,
		      reference_timestamp_delta,
		      kernel_record->file_ID.inode_ID,
		      kernel_record->file_ID.major,
		      kernel_record->file_ID.minor);		      
    break;

  case TAG_ACCEPT:

    result = gzprintf(output_stream,
		      "%c %qx %qx %qx\n",
		      kernel_record->tag,
		      cycle_timestamp_delta,
		      cycle_timestamp_delta,
		      reference_timestamp_delta);
    break;

  default:

    fprintf(stderr, "Per_Task_Writer::append_kernel_record(): ");
    fprintf(stderr,
	    "Invalid tag %c(%hd) for output\n",
	    kernel_record->tag,
	    (unsigned short int)kernel_record->tag);
    exit(1);

  }

  // Ensure that the print succeeded.
  if (result <= 0) {
    fprintf(stderr, "Per_Task_Writer::append_kernel_record(): ");
    fprintf(stderr, "Print failed\n");
    exit(1);
  }

} // Per_Task_Writer::append_kernel_record
// ===================================================================



// ===================================================================
void
Per_Task_Writer::open_output () {

  // It is an error to open an already open output stream.
  if (output_stream != NULL) {
    fprintf(stderr, "Per_Task_Writer::open_output(): ");
    fprintf(stderr, "Stream already open\n");
    exit(1);
  }

  // Attempt to make the directory.
  int mkdir_result = mkdir(output_directory,
			   S_IRUSR | S_IWUSR | S_IXUSR);
  if (mkdir_result == -1) {

    // The directory was not successfully created.  If it already
    // exists, then we can ignore the error.
    if (errno != EEXIST) {
      fprintf(stderr, "Per_Task_Writer::open_output(): ");
      fprintf(stderr, "Failed creation of process directory\n");
      exit(1);
    }

  }

  // Attempt to open the output file.
  int fd = open(output_pathname,
		O_WRONLY | O_CREAT | O_TRUNC,
		S_IRUSR | S_IWUSR);

  // Ensure that it did open.
  if (fd == -1) {
    perror("Per_Task_Writer::open_output(): Failed file open");
    exit(1);
  }

  // Link the file descriptor to a compressed output stream.  Ask that
  // the stream be for (w)riting at the highest (9) compression level
  // accepting (f)iltered data.
  output_stream = gzdopen(fd, "wb9f");

  // Ensure that the linking succeeded.
  if (output_stream == NULL) {
    fprintf(stderr, "Per_Task_Writer::open_output(): ");
    fprintf(stderr, "Failed compression open\n");
    exit(1);
  }

} // Per_Task_Writer::open_output
// ===================================================================
