/**
 * \file Reader.cc
 * \author Scott F. H. Kaplan <sfkaplan@cs.amherst.edu>
 * \date June 2004
 * \brief Header of base class for reading and parsing reference and
 *        kernel traces.
 *
 * A class that defines the interface for reading and parsing
 * reference and kernel traces.  This is an ABSTRACT CLASS: It defines
 * neither the read_reference_record() nor read_kernel_record()
 * methods.  By defining these methods, a subclass actually read and
 * parse some specific trace format and place the needed results into
 * the structure that each method accepts as an argument.
 **/



// ===================================================================
// INCLUDES

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/errno.h>
#include "Reader.hh"
#include "types-and-constants.hh"
// ===================================================================



// ===================================================================
Reader::Reader
(const char* const reference_trace_pathname,
 const char* const kernel_trace_pathname) {

  // Attempt to open the trace file.
  reference_stream = fopen(reference_trace_pathname, "r");

  // Did the reference trace open correctly?
  if (reference_stream == NULL) {

    // No.  Emit the error and exit.
    perror("Failed opening reference trace");
    exit(1);

  }

  // Attempt to open the trace file.
  kernel_stream = fopen(kernel_trace_pathname, "r");

  // Did the kernel trace open correctly?
  if (kernel_stream == NULL) {

    // No.  Emit the error and exit.
    perror("Failed opening kernel trace");
    exit(1);

  }

  // Initialize the record count to zero for each trace.
  reference_record_number = 0;
  kernel_record_number = 0;

  // Allocate space for the buffer.
  buffer = new char[BUFFER_SIZE];
  if (buffer == NULL) {
    fprintf(stderr, "Reader::Reader(): Could not allocate buffer\n");
    exit(1);
  }

} // Reader
// ===================================================================



// ===================================================================
Reader::~Reader () {

  // Attempt to close the reference stream, ensuring success.
  assert(reference_stream != NULL);
  int result = fclose(reference_stream);
  if (result == EOF) {
    perror("Failed closing reference trace");
  }

  // Attempt to close the kernel stream, ensuring success.
  assert(kernel_stream != NULL);
  result = fclose(kernel_stream);
  if (result == EOF) {
    perror("Failed closing kernel trace");
  }

} // ~Reader
// ===================================================================
