/**
 * \file Reader.hh
 * \author Scott F. H. Kaplan <sfkaplan@cs.amherst.edu>
 * \date June 2004
 * \brief Header of base class for reading and parsing reference and
 *        kernel traces.
 *
 * A class that defines the interface for reading and parsing
 * reference and kernel traces.  This is an ABSTRACT CLASS: It defines
 * neither the read_reference_record() nor read_kernel_record()
 * methods.  By defining these methods, a subclass actually read and
 * parse some specific trace format and place the needed results into
 * the structure that each method accepts as an argument.
 */



// ===================================================================
// Avoid multiple inclusion.
#if !defined (_READER_HH)
#define _READER_HH
// ===================================================================



// ===================================================================
// INCLUDES

#include <stdio.h>
#include "types-and-constants.hh"
// ===================================================================




// ===================================================================
// CLASS DEFINITION

/**
 * \brief An abstract class whose subclasses can read different trace
 *        formats.
 *
 * This abstract class is a simple one.  It will open and close the
 * file, and provide the ability to rewind that file.  It does not
 * define how the file is read, but rather provides the method that a
 * subclass must define in order to read a specific reference trace
 * format.
 **/
class Reader {

public: // CONSTRUCTOR & DESTRUCTOR

  /**
   * \brief The constructor.
   * \param reference_trace_pathname The pathname of the reference
   *        trace to be read.
   * \param kernel_trace_pathname The pathname of the kernel trace to
   *        be read.
   *
   * Set up a Reader object by opening the trace files and holding a
   * stream for each.
   **/
  Reader (const char* const reference_trace_pathname,
	  const char* const kernel_trace_pathname);

  /**
   * \brief The deconstructor.
   *
   * Deconstruct a Reader.  Close the trace streams.
   **/
  virtual ~Reader ();

public: // PURE VIRTUAL METHODS

  /**
   * \brief Read the next reference record.
   * \see reference_record_s
   * \param reference_record A structure that will hold the
   *        needed data from the reference record.
   *
   * This pure virtual method must be defined in a subclass so that it
   * reads the next reference record, parses it, and places the
   * required information into the structure passed to it.  Any errors
   * in reading (I/O errors or premature EOF) are expected to abort
   * processing.
   **/
  virtual void read_reference_record
  (reference_record_s* const reference_record) = 0;

  /**
   * \brief Read the next kernel record.
   * \see kernel_record_s
   * \param kernel_record A structure that will hold the needed data
   *        from the kernel record. 
   *
   * This pure virtual method must be defined in a subclass so that it
   * reads the next kernel record, parses it, and places the required
   * information into the structure passed to it.  Any errors in
   * reading (I/O errors or premature EOF) are expected to abort
   * processing.
   **/
  virtual void read_kernel_record
  (kernel_record_s* const kernel_record) = 0;

protected: // DATA

  /** \brief The reference trace input stream. */
  FILE* reference_stream;

  /** \brief The kernel trace input stream. */
  FILE* kernel_stream;

  /** \brief The number of reference records consumed so far. */
  unsigned long long int reference_record_number;

  /** \brief The number of kernel records consumed so far. */
  unsigned long long int kernel_record_number;

  /** \brief A buffer used for input and parsing. */
  char* buffer;

}; // class Reader
// ===================================================================



// ===================================================================
#endif // _READER_HH
// ===================================================================
