/**
 * \file V0_4_Reader.hh
 * \author Scott F. H. Kaplan <sfkaplan@cs.amherst.edu>
 * \date June 2004
 * \brief Header of a class that reads kVMTrace v0.4 traces.
 *
 * A subclass of Reader that can read the reference and kernel traces
 * produced by kVMTrace v0.4.  The data from each record is then
 * transformed appropriately for use the merger utility.  See the
 * README file for more details on the reference trace format.
 */



// ===================================================================
// Avoid multiple inclusion.
#if !defined (_V0_4_READER_HH)
#define _V0_4_READER_HH
// ===================================================================



// ===================================================================
// INCLUDES

#include <stdio.h>
#include "Reader.hh"
// ===================================================================



// ===================================================================
// CLASS DEFINITION

/**
 * The V0_4_Reader definition.  This subclass of Reader that reads
 * kVMTrace v0.4 reference and kernel records.  It defines the
 * read_reference_record() and read_kernel_record() methods that
 * perform the actual reading, parsing, and processing of each record
 * before it is provided to the caller in a provided structure.
 **/
class V0_4_Reader : public Reader {

public: // CONSTRUCTOR

  /**
   * \brief The constructor.
   * \param reference_trace_pathname The pathname of the reference
   *        trace to be read.
   * \param kernel_trace_pathname The pathname of the kernel trace to
   *        be read. 
   * 
   * Set up a V0_4_Reader by initializing its cycle and reference
   * clocks and by opening the trace files and holding a streams for
   * each.
   **/
  V0_4_Reader (const char* const reference_trace_pathname,
	       const char* const kernel_trace_pathname);

public: // VIRTUAL METHODS

  /**
   * \brief Read the next record from the reference trace.
   * \see reference_record_s
   * \param reference_record A structure that will hold the needed
   *        data from the trace record.
   *
   * Read a single reference record, parse its data, process the
   * values, and place the needed results in the provided structure.
   * Any errors in reading (I/O errors or premature EOF) are expected
   * to abort processing.
   **/
  virtual void read_reference_record
  (reference_record_s* const reference_record);

  /**
   * \brief Read the next record from the kernel trace.
   * \see kernel_record_s
   * \param kernel_record A structure that will hold the needed
   *        data from the trace record.
   *
   * Read a single kernel record, parse its data, process the values,
   * and place the needed results in the provided structure.  Any
   * errors in reading (I/O errors or premature EOF) are expected to
   * abort processing.
   **/
  virtual void read_kernel_record
  (kernel_record_s* const kernel_record);

protected: // DATA

  /** \brief The previous CPU cycle time. */
  timestamp_t previous_cycle_time;

  /** \brief The previous reference time. */
  timestamp_t previous_reference_time;

  /** \brief The previous context ID. */
  context_ID_t previous_context_ID;

}; // class V0_4_Reader
// ===================================================================



// ===================================================================
#endif // _V0_4_READER_HH
// ===================================================================
